# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary affiliations."""

from invenio_records_resources.factories.factory import RecordTypeFactory

from ...records.pidprovider import PIDProviderFactory, VocabularyIdProvider
from ...records.systemfields import BaseVocabularyPIDFieldContext
from ...services.permissions import PermissionPolicy
from .config import AffiliationsSearchOptions, service_components
from .schema import AffiliationSchema

record_type = RecordTypeFactory(
    "Affiliation",
    # Data layer
    pid_field_kwargs={
        "create": False,
        "provider": PIDProviderFactory.create(pid_type='aff'),
        "context_cls": BaseVocabularyPIDFieldContext,
    },
    schema_version="1.0.0",
    schema_path="local://affiliations/affiliation-v1.0.0.json",
    # Service layer
    service_schema=AffiliationSchema,
    search_options=AffiliationsSearchOptions,
    service_components=service_components,
    permission_policy_cls=PermissionPolicy,
    # Resource layer
    endpoint_route='/affiliations',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Localization serializer for Vocabularies."""

from functools import partial

from flask import current_app
from flask_babelex import get_locale
from marshmallow import Schema, fields
from marshmallow_utils.fields import BabelGettextDictField


def current_default_locale():
    """Get the Flask app's default locale."""
    if current_app:
        return current_app.config.get('BABEL_DEFAULT_LOCALE', 'en')
    # Use english by default if not specified
    return 'en'


L10NString = partial(BabelGettextDictField, get_locale, current_default_locale)


class VocabularyL10NItemSchema(Schema):
    """Vocabulary serializer schema."""

    id = fields.String(dump_only=True)
    title = L10NString(data_key='title_l10n')
    description = L10NString(data_key='description_l10n')
    props = fields.Dict(dump_only=True)
    icon = fields.String(dump_only=True)
    tags = fields.List(fields.Str(), dump_only=True)


class VocabularyL10NListSchema(Schema):
    """Vocabulary serializer schema."""

    hits = fields.Method('get_hits')

    def get_hits(self, obj_list):
        """Apply hits transformation."""
        schema = self.context['schema_cls']()
        obj_list['hits']['hits'] = [
            schema.dump(h) for h in obj_list['hits']['hits']
        ]
        return obj_list['hits']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Example of a permission policy."""

from invenio_records_permissions import RecordPermissionPolicy
from invenio_records_permissions.generators import AnyUser


class PermissionPolicy(RecordPermissionPolicy):
    """Mock permission policy. All actions allowed."""

    can_search = [AnyUser()]
    can_create = [AnyUser()]
    can_read = [AnyUser()]
    can_update = [AnyUser()]
    can_delete = [AnyUser()]
    can_read_files = [AnyUser()]
    can_update_files = [AnyUser()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary models."""

from invenio_db import db
from invenio_records.models import RecordMetadataBase


class VocabularyType(db.Model):
    """Vocabulary type model."""

    __tablename__ = "vocabularies_types"

    id = db.Column(db.String, primary_key=True)
    pid_type = db.Column(db.String, unique=True)

    @classmethod
    def create(cls, **data):
        """Create a new vocabulary type."""
        with db.session.begin_nested():
            obj = cls(**data)
            db.session.add(obj)
        return obj

    @classmethod
    def dump_obj(cls, field, record, obj):
        """Serializer the object into a record."""
        record[field.attr_name] = {
            'id': obj.id,
            'pid_type': obj.pid_type,
        }

    @classmethod
    def load_obj(cls, field, record):
        """Deserializer the object from a record."""
        data = record.get(field.attr_name)
        if data:
            obj = cls(
                id=data.get('id'),
                pid_type=data.get('pid_type'),
            )
            return obj
        return None


class VocabularyMetadata(db.Model, RecordMetadataBase):
    """Model for vocabulary metadata."""

    __tablename__ = "vocabularies_metadata"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Services module."""

from .service import VocabulariesService, VocabulariesServiceConfig

__all__ = (
    "VocabulariesService",
    "VocabulariesServiceConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from flask_principal import Identity, Need, UserNeed
from invenio_access.permissions import any_user, system_process
from invenio_app.factory import create_api as _create_api
from invenio_cache import current_cache

from invenio_vocabularies.records.api import Vocabulary
from invenio_vocabularies.records.models import VocabularyType

pytest_plugins = ("celery.contrib.pytest", )


@pytest.fixture(scope="module")
def h():
    """Accept JSON headers."""
    return {"accept": "application/json"}


@pytest.fixture(scope="module")
def extra_entry_points():
    """Extra entry points to load the mock_module features."""
    return {
        'invenio_db.models': [
            'mock_module = mock_module.models',
        ],
        'invenio_jsonschemas.schemas': [
            'mock_module = mock_module.jsonschemas',
        ],
        'invenio_search.mappings': [
            'records = mock_module.mappings',
        ]
    }


@pytest.fixture(scope='module')
def app_config(app_config):
    """Mimic an instance's configuration."""
    app_config["JSONSCHEMAS_HOST"] = 'localhost'
    app_config["BABEL_DEFAULT_LOCALE"] = 'en'
    app_config["I18N_LANGUAGES"] = [('da', 'Danish')]
    app_config['RECORDS_REFRESOLVER_CLS'] = \
        "invenio_records.resolver.InvenioRefResolver"
    app_config['RECORDS_REFRESOLVER_STORE'] = \
        "invenio_jsonschemas.proxies.current_refresolver_store"
    return app_config


@pytest.fixture(scope="module")
def create_app(instance_path, entry_points):
    """Application factory fixture."""
    return _create_api


@pytest.fixture(scope="module")
def identity_simple():
    """Simple identity fixture."""
    i = Identity(1)
    i.provides.add(UserNeed(1))
    i.provides.add(Need(method="system_role", value="any_user"))
    return i


@pytest.fixture(scope='module')
def identity():
    """Simple identity to interact with the service."""
    i = Identity(1)
    i.provides.add(UserNeed(1))
    i.provides.add(any_user)
    i.provides.add(system_process)
    return i


@pytest.fixture(scope='module')
def service(app):
    """Vocabularies service object."""
    return app.extensions['invenio-vocabularies'].service


@pytest.fixture()
def lang_type(db):
    """Get a language vocabulary type."""
    return VocabularyType.create(id='languages', pid_type='lng')


@pytest.fixture(scope='function')
def lang_data():
    """Example data."""
    return {
        'id': 'eng',
        'title': {'en': 'English', 'da': 'Engelsk'},
        'description': {
            'en': 'English description',
            'da': 'Engelsk beskrivelse'
        },
        'icon': 'file-o',
        'props': {
            'akey': 'avalue',
        },
        'tags': ['recommended'],
        'type': 'languages',
    }


@pytest.fixture()
def lang_data2(lang_data):
    """Example data for testing invalid cases."""
    data = dict(lang_data)
    data['id'] = 'new'
    return data


@pytest.fixture()
def example_record(db, identity, service, example_data):
    """Example record."""
    vocabulary_type_languages = VocabularyType(name="languages")
    vocabulary_type_licenses = VocabularyType(name="licenses")
    db.session.add(vocabulary_type_languages)
    db.session.add(vocabulary_type_licenses)
    db.session.commit()

    record = service.create(
        identity=identity,
        data=dict(
            **example_data,
            vocabulary_type_id=vocabulary_type_languages.id
        ),
    )

    Vocabulary.index.refresh()  # Refresh the index
    return record


@pytest.fixture(scope='function')
def lang_data_many(lang_type, lic_type, lang_data, service, identity):
    """Create many language vocabulary."""
    lang_ids = ['fr', 'tr', 'gr', 'ger', 'es']
    data = dict(lang_data)

    for lang_id in lang_ids:
        data['id'] = lang_id
        service.create(identity, data)
    Vocabulary.index.refresh()  # Refresh the index
    return lang_ids


# FIXME: https://github.com/inveniosoftware/pytest-invenio/issues/30
# Without this, success of test depends on the tests order
@pytest.fixture()
def cache():
    """Empty cache."""
    try:
        yield current_cache
    finally:
        current_cache.clear()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2021 Graz University of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Commands to create and manage vocabulary."""

import csv
from os.path import dirname, join

import click
from flask.cli import with_appcontext
from flask_principal import Identity
from invenio_access import any_user
from invenio_db import db

from invenio_vocabularies.contrib.subjects.subjects import subject_record_type
from invenio_vocabularies.records.models import VocabularyType
from invenio_vocabularies.services.service import VocabulariesService

data_directory = join(dirname(__file__), "data")


def get_available_vocabularies():
    """Specify the available vocabularies."""
    return {
        "licenses": {
            "path": join(data_directory, "licenses.csv"),
        },
        "subjects": {
            "path": join(data_directory, "subjects.csv"),
            "specific": _create_subjects_vocabulary,
        },
    }


def _load_csv_data(path):
    with open(path) as f:
        reader = csv.DictReader(f, skipinitialspace=True)
        dicts = [row for row in reader]
        return dicts


def _create_subjects_vocabulary(vocabulary_type_name, source_path):
    identity = Identity(1)
    identity.provides.add(any_user)
    service = subject_record_type.service_cls()

    rows = _load_csv_data(source_path)

    records = []
    for row in rows:
        metadata = {
            "title": row["title"],
            "term": row["id"],
            "identifier": row["id"],
            "scheme": row["scheme"],
        }

        record = service.create(
            identity=identity,
            data={
                "metadata": metadata,
            },
        )

        records.append(record)

    return records


def _create_vocabulary(vocabulary_type_name, source_path):
    identity = Identity(1)
    identity.provides.add(any_user)
    service = VocabulariesService()

    # Load data
    rows = _load_csv_data(source_path)

    # Create vocabulary type
    vocabulary_type = VocabularyType(name=vocabulary_type_name)
    db.session.add(vocabulary_type)
    db.session.commit()

    i18n = ["title", "description"]  # Attributes with i18n support
    other = ["icon"]  # Other top-level attributes

    default_language = "en"  # Static (dependent on the files)

    metadata = {"title": {}, "description": {}, "props": {}}

    records = []
    for row in rows:
        for attribute in row:
            value = row[attribute]
            if attribute in i18n:
                metadata[attribute][default_language] = value
            elif any(map(lambda s: value.startswith(s + "_"), i18n)):
                [prefix_attr, language] = attribute.split("_", 1)
                metadata[prefix_attr][language] = value
            elif attribute in other:
                metadata[attribute] = value
            else:
                metadata["props"][attribute] = value

        # Create record
        record = service.create(
            identity=identity,
            data={
                "metadata": metadata,
                "vocabulary_type_id": vocabulary_type.id,
            },
        )

        records.append(record)

    return records


@click.group()
def vocabularies():
    """Vocabularies command."""
    pass


@vocabularies.command(name="import")
@click.argument(
    "vocabulary_types",
    nargs=-1,
    type=click.Choice([v for v in get_available_vocabularies()]),
)
@with_appcontext
def load(vocabulary_types):
    """Index CSV-based vocabularies in Elasticsearch."""
    click.echo("creating vocabularies...", color="blue")

    for vocabulary_type in vocabulary_types:
        vocabulary = get_available_vocabularies()[vocabulary_type]
        if VocabularyType.query.filter_by(name=vocabulary_type).count() > 0:
            click.echo(
                "vocabulary type {} already exists, skipping".format(
                    vocabulary_type
                ),
                color="red",
            )
            continue

        click.echo(
            "creating vocabulary type {}...".format(vocabulary_type),
            color="blue",
        )

        fun_create_vocabulary = vocabulary.get("specific", _create_vocabulary)
        items = fun_create_vocabulary(vocabulary_type, vocabulary["path"])

        click.echo(
            "created {} vocabulary items successfully".format(len(items)),
            color="green",
        )
    click.echo("vocabularies created", color="green")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resources layer tests."""

import pytest

from invenio_vocabularies.records.api import Vocabulary


#
# Fixtures
#
@pytest.fixture()
def example_data():
    """Example data for records."""
    return [
        {'id': 'cc-by', 'title': {'en': 'Creative Commons Attribution'},
         'type': 'licenses'},
        {'id': 'cc0', 'title': {'en': 'Creative Commons Zero'},
         'type': 'licenses'},
    ]


@pytest.fixture()
def example_records(db, identity, service, example_data):
    """Create some example records."""
    service.create_type(identity, 'licenses', 'lic')
    records = []
    for data in example_data:
        records.append(service.create(identity, data))
    Vocabulary.index.refresh()
    return records


@pytest.fixture()
def prefix():
    """API prefix."""
    return '/vocabularies/licenses/'


#
# Tests
#
def test_get(client, example_records, h, prefix):
    """Test the endpoint to retrieve a single item."""
    id_ = example_records[0].id

    res = client.get(f'{prefix}{id_}', headers=h)
    assert res.status_code == 200
    assert res.json['id'] == id_
    # Test links
    assert res.json['links'] == {
        'self': 'https://127.0.0.1:5000/api/vocabularies/licenses/cc-by'
    }


def test_not_found(client, example_records, h, prefix):
    """Test not found."""
    id_ = example_records[0].id
    # invalid id
    res = client.get(f'{prefix}invalid-id', headers=h,)
    assert res.status_code == 404
    # invalid type (wrong type for pid)
    res = client.get(f'/vocabularies/invalid/{id_}', headers=h,)
    assert res.status_code == 404
    # trailing slash on search
    res = client.get(f'{prefix}/', headers=h,)
    assert res.status_code == 404
    # invalid type (not existing)
    res = client.get(f'/vocabularies/invalid', headers=h,)
    assert res.status_code == 404



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary subjects."""

from invenio_records_resources.factories.factory import RecordTypeFactory

from invenio_vocabularies.contrib.subjects.permissions import PermissionPolicy
from invenio_vocabularies.contrib.subjects.schema import SubjectSchema

subject_record_type = RecordTypeFactory(
    "Subject", SubjectSchema,
    permission_policy_cls=PermissionPolicy,
    endpoint_route='/subjects'
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects schema."""

from marshmallow import EXCLUDE, INCLUDE, Schema, fields, validate
from marshmallow_utils.fields import Links


class MetadataSchema(Schema):
    """Basic metadata schema class."""

    class Meta:
        """Meta class to accept unknown fields."""

        unknown = INCLUDE

    scheme = fields.Str(required=True, validate=validate.Length(min=3))
    term = fields.Str(required=True, validate=validate.Length(min=3))
    identifier = fields.Str(required=True, validate=validate.Length(min=3))
    title = fields.Str(required=True, validate=validate.Length(min=3))


class SubjectSchema(Schema):
    """Schema for records v1 in JSON."""

    class Meta:
        """Meta class to reject unknown fields."""

        unknown = EXCLUDE

    id = fields.Str()
    metadata = fields.Nested(MetadataSchema)
    created = fields.Str()
    updated = fields.Str()
    links = Links()
    revision_id = fields.Integer(dump_only=True)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create vocabularies tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import mysql, postgresql

# revision identifiers, used by Alembic.
revision = "4f365fced43f"
down_revision = "8ff82dfb0be8"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        "subject_metadata",
        sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False
        ),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()),
                "postgresql",
            )
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_subject_metadata")),
    )
    op.create_table(
        "vocabularies_metadata",
        sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False
        ),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()),
                "postgresql",
            )
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_vocabularies_metadata")),
    )
    op.create_table(
        "vocabularies_types",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("pid_type", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_vocabularies_types")),
        sa.UniqueConstraint(
            "pid_type", name=op.f("uq_vocabularies_types_pid_type")
        ),
    )


def downgrade():
    """Downgrade database."""
    op.drop_table("vocabularies_types")
    op.drop_table("vocabularies_metadata")
    op.drop_table("subject_metadata")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary affiliations configuration."""

from flask_babelex import lazy_gettext as _
from invenio_records_resources.services import SearchOptions
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.params import \
    SuggestQueryParser

from ...services.components import PIDComponent


class AffiliationsSearchOptions(SearchOptions):
    """Search options."""

    suggest_parser_cls = SuggestQueryParser.factory(
        fields=[
            'name^100',
            'acronym^20',
            'title.*^5',
            'title.*._2gram',
            'title.*._3gram',
        ],
    )

    sort_default = 'bestmatch'

    sort_default_no_query = 'name'

    sort_options = {
        "bestmatch": dict(
            title=_('Best match'),
            fields=['_score'],  # ES defaults to desc on `_score` field
        ),
        "name": dict(
            title=_('Name'),
            fields=['name_sort'],
        ),
        "newest": dict(
            title=_('Newest'),
            fields=['-created'],
        ),
        "oldest": dict(
            title=_('Oldest'),
            fields=['created'],
        ),
    }


service_components = [
    # Order of components are important!
    DataComponent,
    PIDComponent,
]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the vocabulary service."""

import pytest
from invenio_cache import current_cache
from invenio_pidstore.errors import PIDAlreadyExists, PIDDeletedError
from marshmallow.exceptions import ValidationError
from sqlalchemy.exc import IntegrityError

from invenio_vocabularies.records.api import Vocabulary, VocabularyType


#
# Fixtures
#
@pytest.fixture()
def lic_type(db):
    """Get a language vocabulary type."""
    return VocabularyType.create(id='licenses', pid_type='lic')


#
# Tests
#
def test_simple_flow(lang_type, lic_type, lang_data, service, identity):
    """Test a simple vocabulary service flow."""
    # Create it
    item = service.create(identity, lang_data)
    id_ = item.id

    assert item.id == lang_data['id']
    for k, v in lang_data.items():
        assert item.data[k] == v

    # Read it
    read_item = service.read(('languages', 'eng'), identity)
    assert item.id == read_item.id
    assert item.data == read_item.data
    assert read_item.data['type'] == 'languages'

    # Refresh index to make changes live.
    Vocabulary.index.refresh()

    # Search it
    res = service.search(
        identity, type='languages', q=f"id:{id_}", size=25, page=1)
    assert res.total == 1
    assert list(res.hits)[0] == read_item.data

    # Search another type
    res = service.search(
        identity, type='licenses', q=f"id:{id_}", size=25, page=1)
    assert res.total == 0

    # Update it
    data = read_item.data
    data['title']['en'] = 'New title'
    update_item = service.update(('languages', id_), identity, data)
    assert item.id == update_item.id
    assert update_item['title']['en'] == 'New title'

    # Delete it
    assert service.delete(('languages', id_), identity)

    # Refresh to make changes live
    Vocabulary.index.refresh()

    # Fail to retrieve it
    # - db
    pytest.raises(
        PIDDeletedError, service.read, ('languages', id_), identity)
    # - search
    res = service.search(
        identity, type='languages', q=f"id:{id_}", size=25, page=1)
    assert res.total == 0


def test_pid_already_registered(lang_type, lang_data, service, identity):
    """Recreating a record with same id should fail."""
    service.create(identity, lang_data)
    pytest.raises(PIDAlreadyExists, service.create, identity, lang_data)


def test_extra_fields(lang_data2, service, identity):
    """Extra fields in data should fail."""
    lang_data2['invalid'] = 1
    pytest.raises(ValidationError, service.create, identity, lang_data2)


def test_missing_or_invalid_type(lang_data2, service, identity):
    """A missing type should raise validation error."""
    # No type specified
    del lang_data2['type']
    pytest.raises(ValidationError, service.create, identity, lang_data2)

    # Invalid value data types
    for val in [1, {'id': 'languages'}, ]:
        lang_data2['type'] = val
        pytest.raises(
            ValidationError, service.create, identity, lang_data2)

    # Non-existing type
    lang_data2['type'] = 'invalid'
    pytest.raises(
            ValidationError, service.create, identity, lang_data2)


def test_update(lang_type, lang_data2, service, identity):
    """Removing keys should work."""
    lang_data2['id'] = 'dan'
    item = service.create(identity, lang_data2)
    id_ = item.id

    data = item.data
    # Unset a subkey
    del data['title']['en']
    # Unset a top-level key
    del data['icon']
    update_item = service.update(('languages', id_), identity, data)
    # Ensure they really got cleared.
    assert 'en' not in update_item.data['title']
    assert 'icon' not in update_item.data


def test_missing_or_invalid_type(lang_data2, service, identity):
    """Custom props should only accept strings."""
    # Invalid data types
    for v in [1, {'id': 'test'}]:
        lang_data2['props']['newkey'] = v
        pytest.raises(
            ValidationError, service.create, identity, lang_data2)


def test_read_all_no_cache(
    lang_type, lang_data_many, service, identity, cache
):
    """ read_all method should return all languages created in this scope. """
    items = service.read_all(identity, fields=['id'],
                             type='languages', cache=False)
    assert set(lang_data_many).issubset(set([i["id"] for i in items]))
    cached = current_cache.get("id")
    assert not cached


def test_read_all_cache(lang_type, lang_data_many, service, identity, cache):
    """ read_all method should return all languages created in this scope. """
    items = service.read_all(identity, fields=['id'],
                             type='languages', cache=True)
    assert set(lang_data_many).issubset(set([i["id"] for i in items]))
    cached = current_cache.get("languages_id")
    assert cached is not None


def test_read_many(lang_type, lang_data_many, service, identity):
    """ read_many method should return all requested languages. """
    ids_ = ['fr', 'tr', 'es']
    items = service.read_many(identity, type='languages', ids=ids_, fields=[])
    item_ids = [i.id for i in items._results]
    assert {'fr', 'tr', 'es'} == set(item_ids)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.


"""Example of a record API."""

from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2
from invenio_records.dumpers import ElasticsearchDumper
from invenio_records.dumpers.relations import RelationDumperExt
from invenio_records.systemfields import ConstantField, RelationsField
from invenio_records_resources.records.api import Record as RecordBase
from invenio_records_resources.records.systemfields import IndexField, \
    PIDField, PIDListRelation

from invenio_vocabularies.records.api import Vocabulary

from . import models


class Record(RecordBase):
    """Example bibliographic record API."""

    model_cls = models.RecordMetadata
    schema = ConstantField(
        '$schema', 'local://records/record-v1.0.0.json')
    index = IndexField('records-record-v1.0.0', search_alias='records')
    pid = PIDField('id', provider=RecordIdProviderV2)

    # Definitions of relationships from a bibliographic record to the
    # generic vocabularies.
    relations = RelationsField(
        languages=PIDListRelation(
            'metadata.languages',
            attrs=['id', 'title'],
            pid_field=Vocabulary.pid.with_type_ctx('languages')
        ),
    )

    dumper = ElasticsearchDumper(
        extensions=[
            RelationDumperExt('relations'),
        ]
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabularies configuration."""

from .resources.resource import VocabulariesResourceConfig
from .services.service import VocabulariesServiceConfig

VOCABULARIES_RESOURCE_CONFIG = VocabulariesResourceConfig
"""Configure the resource."""

VOCABULARIES_SERVICE_CONFIG = VocabulariesServiceConfig
"""Configure the service."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio vocabularies views."""


def create_blueprint_from_app(app):
    """Create app blueprint."""
    return app.extensions["invenio-vocabularies"].resource.as_blueprint()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""System fields module."""

from .pid import BaseVocabularyPIDFieldContext, VocabularyPIDFieldContext

__all__ = (
    'BaseVocabularyPIDFieldContext',
    'VocabularyPIDFieldContext',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest

from invenio_vocabularies.contrib.affiliations.resources import \
    AffiliationsResource, AffiliationsResourceConfig
from invenio_vocabularies.contrib.affiliations.services import \
    AffiliationsService, AffiliationsServiceConfig


@pytest.fixture(scope="module")
def extra_entry_points():
    """Extra entry points to load the mock_module features."""
    return {
        "invenio_db.models": [
            "affiliations = invenio_vocabularies.contrib.affiliations.models",
        ],
        "invenio_jsonschemas.schemas": [
            "affiliations = \
                invenio_vocabularies.contrib.affiliations.jsonschemas",
        ],
        "invenio_search.mappings": [
            "affiliations = \
                invenio_vocabularies.contrib.affiliations.mappings",
        ]
    }


@pytest.fixture(scope="function")
def affiliation_full_data():
    """Full affiliation data."""
    return {
        "acronym": "TEST",
        "id": "cern",
        "identifiers": [
            {"identifier": "03yrm5c26", "scheme": "ror"}
        ],
        "name": "Test affiliation",
        "title": {
            "en": "Test affiliation",
            "es": "Afiliacion de test"
        }
    }


@pytest.fixture(scope='module')
def service():
    """Affiliations service object."""
    return AffiliationsService(config=AffiliationsServiceConfig)


@pytest.fixture(scope="module")
def resource(service):
    """Affiliations resource object."""
    return AffiliationsResource(AffiliationsResourceConfig, service)


@pytest.fixture(scope="module")
def base_app(base_app, resource, service):
    """Application factory fixture.

    Registers affiliations' resource and service.
    """
    base_app.register_blueprint(resource.as_blueprint())
    registry = base_app.extensions['invenio-records-resources'].registry
    registry.register(service, service_id='affiliations-service')
    yield base_app



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Proxies to the service and resource."""

from flask import current_app
from werkzeug.local import LocalProxy


def _ext_proxy(attr):
    return LocalProxy(
        lambda: getattr(current_app.extensions['invenio-vocabularies'], attr))


current_service = _ext_proxy('service')
"""Proxy to the instantiated vocabulary service."""


current_resource = _ext_proxy('resource')
"""Proxy to the instantiated vocabulary resource."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Example of a record model."""

from invenio_db import db
from invenio_records.models import RecordMetadataBase


class RecordMetadata(db.Model, RecordMetadataBase):
    """Model for mock module metadata."""

    __tablename__ = 'mock_metadata'



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the affiliation vocabulary service."""

import pytest
from invenio_pidstore.errors import PIDAlreadyExists, PIDDeletedError
from marshmallow.exceptions import ValidationError
from sqlalchemy.exc import IntegrityError

from invenio_vocabularies.contrib.affiliations.api import Affiliations


def test_simple_flow(app, db, service, identity, affiliation_full_data):
    """Test a simple vocabulary service flow."""
    # Create it
    item = service.create(identity, affiliation_full_data)
    id_ = item.id

    assert item.id == affiliation_full_data['id']
    for k, v in affiliation_full_data.items():
        assert item.data[k] == v

    # Read it
    read_item = service.read('cern', identity)
    assert item.id == read_item.id
    assert item.data == read_item.data

    # Refresh index to make changes live.
    Affiliations.index.refresh()

    # Search it
    res = service.search(
        identity, q=f"id:{id_}", size=25, page=1)
    assert res.total == 1
    assert list(res.hits)[0] == read_item.data

    # Update it
    data = read_item.data
    data['title']['en'] = 'New title'
    update_item = service.update(id_, identity, data)
    assert item.id == update_item.id
    assert update_item['title']['en'] == 'New title'

    # Delete it
    assert service.delete(id_, identity)

    # Refresh to make changes live
    Affiliations.index.refresh()

    # Fail to retrieve it
    # - db
    pytest.raises(
        PIDDeletedError, service.read, id_, identity)
    # - search
    res = service.search(
        identity, q=f"id:{id_}", size=25, page=1)
    assert res.total == 0


def test_pid_already_registered(
    app, db, service, identity, affiliation_full_data
):
    """Recreating a record with same id should fail."""
    service.create(identity, affiliation_full_data)
    pytest.raises(
        PIDAlreadyExists, service.create, identity, affiliation_full_data)


def test_extra_fields(app, db, service, identity, affiliation_full_data):
    """Extra fields in data should fail."""
    affiliation_full_data['invalid'] = 1
    pytest.raises(
        ValidationError, service.create, identity, affiliation_full_data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module for managing vocabularies."""

from . import config
from .resources.resource import VocabulariesResource
from .services.service import VocabulariesService


class InvenioVocabularies(object):
    """Invenio-Vocabularies extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        self.resource = None
        self.service = None
        self.subjects_service = None
        self.subjects_resource = None
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        self.init_resource(app)
        app.extensions["invenio-vocabularies"] = self

    def init_resource(self, app):
        """Initialize vocabulary resources."""
        self.service = VocabulariesService(
            config=app.config["VOCABULARIES_SERVICE_CONFIG"],
        )
        self.resource = VocabulariesResource(
            service=self.service,
            config=app.config["VOCABULARIES_RESOURCE_CONFIG"],
        )

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("VOCABULARIES_"):
                app.config.setdefault(k, getattr(config, k))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test alembic recipes for Invenio-RDM-Records."""

import pytest
from invenio_db.utils import drop_alembic_version_table


def assert_alembic(alembic, table_excludes):
    """Assert that metadata of alembic and db matches.

    This method allows omitting tables dynamically created for tests.
    """
    assert not list(filter(
        # x[0] is the operation and x[1] is the table
        lambda x: x[0] == 'add_table' and x[1].name not in table_excludes,
        alembic.compare_metadata()
    ))


def test_alembic(app, database):
    """Test alembic recipes."""
    db = database
    ext = app.extensions['invenio-db']

    if db.engine.name == 'sqlite':
        raise pytest.skip('Upgrades are not supported on SQLite.')

    # Check that this package's SQLAlchemy models have been properly registered
    tables = [x.name for x in db.get_tables_for_bind()]
    assert 'vocabularies_metadata' in tables
    assert 'vocabularies_types' in tables
    assert 'subject_metadata' in tables

    # Check that Alembic agrees that there's no further tables to create.
    assert_alembic(ext.alembic, ['mock_metadata'])

    # Drop everything and recreate tables all with Alembic
    db.drop_all()
    drop_alembic_version_table()
    ext.alembic.upgrade()
    assert_alembic(ext.alembic, ['mock_metadata'])

    # Try to upgrade and downgrade
    ext.alembic.stamp()
    ext.alembic.downgrade(target='96e796392533')
    ext.alembic.upgrade()
    assert_alembic(ext.alembic, ['mock_metadata'])

    # Cleanup
    drop_alembic_version_table()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resources layer tests."""

import pytest

from invenio_vocabularies.records.api import Vocabulary


#
# Fixtures
#
@pytest.fixture(scope='module')
def example_data():
    """Example data for records."""
    return {
        'id': 'text',
        'title': {
            'en': 'Text',
            'da': 'Tekst',
        },
        'description': {
            'en': 'Publications',
            'da': 'Publikationer',
        },
        'icon': 'file-o',
        'props': {
        },
        'type': 'resourcetypes2'
    }


@pytest.fixture(scope='module')
def example_record(database, identity, service, example_data):
    """Create a vocabulary record."""
    service.create_type(identity, 'resourcetypes2', 'rt2')
    item = service.create(identity, example_data)
    Vocabulary.index.refresh()
    return item


@pytest.fixture()
def prefix():
    """API prefix."""
    return '/vocabularies/resourcetypes2'


@pytest.fixture()
def h():
    """Header for localised versions."""
    return {'accept': 'application/vnd.inveniordm.v1+json'}


@pytest.fixture(scope='module')
def expected_da():
    """Expected serialization when danish chosen."""
    return {
        'id': 'text',
        'title_l10n': 'Tekst',
        'description_l10n': 'Publikationer',
        'icon': 'file-o',
        'props': {},
    }


@pytest.fixture(scope='module')
def expected_en():
    """Expected serialization when english chosen."""
    return {
        'id': 'text',
        'title_l10n': 'Text',
        'description_l10n': 'Publications',
        'icon': 'file-o',
        'props': {},
    }


#
# Tests
#
def test_get(client, example_record, h, prefix, expected_da, expected_en):
    """Test the endpoint to retrieve a single item."""
    id_ = example_record.id

    # Default locale
    res = client.get(f'{prefix}/{id_}', headers=h)
    assert res.json == expected_en

    # Choose via querystring (?ln=da)
    res = client.get(f'{prefix}/{id_}?ln=da', headers=h)
    assert res.json == expected_da
    res = client.get(f'{prefix}/{id_}?ln=en', headers=h)
    assert res.json == expected_en

    # Choose via header
    h['accept-language'] = 'da'
    res = client.get(f'{prefix}/{id_}', headers=h)
    assert res.json == expected_da
    h['accept-language'] = 'en'
    res = client.get(f'{prefix}/{id_}', headers=h)
    assert res.json == expected_en


def test_search(client, example_record, h, prefix, expected_da, expected_en):
    """Test search result serialization."""
    expected_en = {'hits': {'hits': [expected_en], 'total': 1}}
    expected_da = {'hits': {'hits': [expected_da], 'total': 1}}

    # Default locale
    res = client.get(f'{prefix}', headers=h)
    assert res.json == expected_en

    # Choose via querystring (?ln=da)
    res = client.get(f'{prefix}?ln=da', headers=h)
    assert res.json == expected_da
    res = client.get(f'{prefix}?ln=en', headers=h)
    assert res.json == expected_en

    # Choose via header
    h['accept-language'] = 'da'
    res = client.get(f'{prefix}', headers=h)
    assert res.json == expected_da
    h['accept-language'] = 'en'
    res = client.get(f'{prefix}', headers=h)
    assert res.json == expected_en



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary service."""

from elasticsearch_dsl.query import Q
from elasticsearch_dsl.response import Response
from flask_babelex import lazy_gettext as _
from invenio_cache import current_cache
from invenio_db import db
from invenio_records_resources.services import Link, LinksTemplate, \
    RecordService, RecordServiceConfig, SearchOptions, pagination_links
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.params import FilterParam, \
    SuggestQueryParser

from ..records.api import Vocabulary
from ..records.models import VocabularyType
from .components import PIDComponent, VocabularyTypeComponent
from .permissions import PermissionPolicy
from .schema import VocabularySchema


class VocabularySearchOptions(SearchOptions):
    """Search options."""

    params_interpreters_cls = [
        FilterParam.factory(param='tags', field='tags'),
    ] + SearchOptions.params_interpreters_cls

    suggest_parser_cls = SuggestQueryParser.factory(
        fields=[
            'id.text^100',
            'id.text._2gram',
            'id.text._3gram',
            'title.en^5',
            'title.en._2gram',
            'title.en._3gram',
        ],
    )

    sort_default = 'bestmatch'

    sort_default_no_query = 'title'

    sort_options = {
        "bestmatch": dict(
            title=_('Best match'),
            fields=['_score'],  # ES defaults to desc on `_score` field
        ),
        "title": dict(
            title=_('Title'),
            fields=['title_sort'],
        ),
        "newest": dict(
            title=_('Newest'),
            fields=['-created'],
        ),
        "oldest": dict(
            title=_('Oldest'),
            fields=['created'],
        ),
    }


class VocabulariesServiceConfig(RecordServiceConfig):
    """Vocabulary service configuration."""

    permission_policy_cls = PermissionPolicy
    record_cls = Vocabulary
    schema = VocabularySchema

    search = VocabularySearchOptions

    components = [
        # Order of components are important!
        VocabularyTypeComponent,
        DataComponent,
        PIDComponent,
    ]

    links_item = {
        "self": Link(
            "{+api}/vocabularies/{type}/{id}",
            vars=lambda record, vars: vars.update({
                "id": record.pid.pid_value,
                "type": record.type.id,
            })
        ),
    }

    links_search = pagination_links("{+api}/vocabularies/{type}{?args*}")


class VocabulariesService(RecordService):
    """Vocabulary service."""

    def create_type(self, identity, id, pid_type):
        """Create a new vocabulary type."""
        self.require_permission(identity, "manage")
        type_ = VocabularyType.create(id=id, pid_type=pid_type)
        db.session.commit()
        return type_

    def search(self, identity, params=None, es_preference=None, type=None,
               **kwargs):
        """Search for vocabulary entries."""
        self.require_permission(identity, 'search')

        # If not found, NoResultFound is raised (caught by the resource).
        vocabulary_type = VocabularyType.query.filter_by(id=type).one()

        # Prepare and execute the search
        params = params or {}
        search_result = self._search(
            'search',
            identity,
            params,
            es_preference,
            extra_filter=Q('term', type__id=vocabulary_type.id),
            **kwargs
        ).execute()

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=LinksTemplate(self.config.links_search, context={
                "args": params,
                "type": vocabulary_type.id,
            }),
            links_item_tpl=self.links_item_tpl,
        )

    def read_all(self, identity, fields, type, cache=True, **kwargs):
        """Search for records matching the querystring."""
        cache_key = type + "_" + "-".join(fields)
        results = current_cache.get(cache_key)
        es_query = Q("match_all")

        if not results:
            # If not found, NoResultFound is raised (caught by the resource).
            vocabulary_type = VocabularyType.query.filter_by(id=type).one()
            vocab_id_filter = Q('term', type__id=vocabulary_type.id)

            results = self._read_many(
                identity, es_query, fields,
                extra_filter=vocab_id_filter, **kwargs)
            if cache:
                # ES DSL Response is not pickable.
                # If saved in cache serialization wont work with to_dict()
                current_cache.set(cache_key, results.to_dict())

        else:
            search = self.create_search(
                identity=identity,
                record_cls=self.record_cls,
                search_opts=self.config.search,
                permission_action='search',
            ).query(es_query)

            results = Response(search, results)

        return self.result_list(self, identity, results)

    def read_many(self, identity, type, ids, fields=None, **kwargs):
        """Search for records matching the querystring filtered by ids."""
        es_query = Q("match_all")
        vocabulary_type = VocabularyType.query.filter_by(id=type).one()
        vocab_id_filter = Q('term', type__id=vocabulary_type.id)
        filters = []

        for id_ in ids:
            filters.append(Q('term', **{"id": id_}))
        filter = Q('bool', minimum_should_match=1, should=filters)
        filter = filter & vocab_id_filter
        results = self._read_many(
            identity, es_query, fields,
            extra_filter=filter, **kwargs)

        return self.result_list(self, identity, results)



#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create affiliations table."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import mysql, postgresql

# revision identifiers, used by Alembic.
revision = '6312f33645c1'
down_revision = '4f365fced43f'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        'affiliation_metadata',
        sa.Column(
            'created',
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), 'mysql'),
            nullable=False
        ),
        sa.Column(
            'updated',
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), 'mysql'),
            nullable=False
        ),
        sa.Column(
            'id',
            sqlalchemy_utils.types.uuid.UUIDType(),
            nullable=False
        ),
        sa.Column(
            'json',
            sa.JSON()
            .with_variant(sqlalchemy_utils.types.json.JSONType(), 'mysql')
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()),
                'postgresql')
            .with_variant(sqlalchemy_utils.types.json.JSONType(), 'sqlite'),
            nullable=True
        ),
        sa.Column('version_id', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_affiliation_metadata'))
    )


def downgrade():
    """Downgrade database."""
    op.drop_table('affiliation_metadata')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data access layer tests."""

import pytest
from jsonschema import ValidationError as SchemaValidationError
from sqlalchemy import inspect

from invenio_vocabularies.records.api import Vocabulary


def test_record_schema_validation(app, db, lang_type):
    """Record schema validation."""
    # Good data
    record = Vocabulary.create({
        'id': 'eng',
        'title': {'en': 'English', 'da': 'Engelsk'},
        'description': {'en': 'English', 'da': 'Engelsk'},
        'icon': 'en',
        'props': {
            'akey': 'a value'
        },
        'tags': ['recommended'],
    }, type=lang_type)
    assert record.schema

    # Bad data
    examples = [
        # title/descriptions are objects of key/string.
        {'id': 'en', 'title': 'not a string'},
        {'id': 'en', 'title': {'en': 123}},
        {'id': 'en', 'description': 'not a string'},
        # icon must be strings
        {'id': 'en', 'icon': 123},
        # props values must be strings
        {'id': 'en', 'props': {'key': 123}},
        {'id': 'en', 'props': {'key': {'test': 'test'}}},
        # Additional properties false
        {'id': 'en', 'metadata': {'title': 'test'}},
    ]

    for ex in examples:
        pytest.raises(SchemaValidationError, Vocabulary.create, ex)


def test_record_indexing(app, db, es, example_record, indexer, search_get):
    """Test indexing of a record."""
    # Index document in ES
    assert indexer.index(example_record)["result"] == "created"

    # Retrieve document from ES
    data = search_get(id=example_record.id)

    # Loads the ES data and compare
    record = Vocabulary.loads(data["_source"])
    assert record == example_record
    assert record.id == example_record.id
    assert record.revision_id == example_record.revision_id
    assert record.created == example_record.created
    assert record.updated == example_record.updated

    # Check system fields - i.e reading related type object from
    assert record == example_record
    assert record.type.id == 'languages'
    assert record.type.pid_type == 'lng'

    # Check that object was recrated without hitting DB
    assert inspect(record.type).persistent is False
    Vocabulary.type.session_merge(record)
    assert inspect(record.type).persistent is True


def test_record_pids(app, db, lang_type, lic_type):
    """Test record pid creation."""
    record = Vocabulary.create({
        'id': 'eng', 'title': {'en': 'English', 'da': 'Engelsk'}},
        type=lang_type
    )
    Vocabulary.pid.create(record)
    assert record.type == lang_type
    assert record.pid.pid_value == 'eng'
    assert record.pid.pid_type == 'lng'
    assert Vocabulary.pid.resolve(('languages', 'eng'))

    record = Vocabulary.create({
        'id': 'cc-by', 'title': {'en': 'CC-BY', 'da': 'CC-BY'}
    }, type=lic_type)
    Vocabulary.pid.create(record)
    assert record.type == lic_type
    assert record.pid.pid_value == 'cc-by'
    assert record.pid.pid_type == 'lic'
    assert Vocabulary.pid.resolve(('licenses', 'cc-by'))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""REST API search tests."""

import pytest

from invenio_vocabularies.records.api import Vocabulary


#
# Fixtures
#
@pytest.fixture(scope='module')
def example_data():
    """Example data for records."""
    return [
        {'id': 'text', 'title': {'en': 'Text'}, 'type': 'resourcetypes'},
        {'id': 'data', 'title': {'en': 'Data'}, 'type': 'resourcetypes',
         'tags': ['recommended']},
    ]


@pytest.fixture(scope='module')
def example_records(database, identity, service, example_data):
    """Create some example records."""
    service.create_type(identity, 'resourcetypes', 'rt')
    records = []
    for data in example_data:
        records.append(service.create(identity, data))
    Vocabulary.index.refresh()
    return records


@pytest.fixture()
def prefix():
    """API prefix."""
    return '/vocabularies/resourcetypes'


#
# Tests
#
def test_invalid_type(app, client, h):
    """Test invalid type."""
    # invalid type
    res = client.get('/vocabularies/invalid', headers=h)
    assert res.status_code == 404
    # trailing slash
    res = client.get('/vocabularies/invalid/', headers=h)
    assert res.status_code == 404


def test_type_as_query_arg(app, lang_type, example_records, client, h, prefix):
    """Test passing an invalid query parameter."""
    # Ensure we cannot pass invalid query parameters
    res = client.get(f'{prefix}?invalid=test', headers=h)
    assert res.status_code == 200
    assert 'invalid' not in res.json['links']['self']

    # It should not be possible to pass 'type' in query args (because it's in
    # the URL route instead)
    res = client.get(f'{prefix}?type={lang_type.id}', headers=h)
    assert res.status_code == 200
    assert lang_type.id not in res.json['links']['self']
    assert 'type=' not in res.json['links']['self']
    assert 'resourcetypes' in res.json['links']['self']


def test_search(client, example_records, h, prefix):
    """Test a successful search."""
    res = client.get(prefix, headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 2
    assert res.json['sortBy'] == 'title'


def test_query_q(client, example_records, h, prefix):
    """Test a successful search."""
    # Test query (q=)
    res = client.get(f'{prefix}?q=title.en:text', headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json['sortBy'] == 'bestmatch'

    # Test sort
    res = client.get(f'{prefix}?q=*&sort=bestmatch', headers=h)
    assert res.status_code == 200
    assert res.json['sortBy'] == 'bestmatch'

    # Test size
    res = client.get(f'{prefix}?size=1&page=1', headers=h)
    assert res.status_code == 200
    assert res.json['hits']['total'] == 2
    assert len(res.json['hits']['hits']) == 1
    assert 'next' in res.json['links']


def test_tags_filter(client, example_records, h, prefix):
    """Test filter on tags."""
    # Test query (q=)
    res = client.get(f'{prefix}?tags=recommended', headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations schema."""

from invenio_records_resources.services.records.schema import BaseRecordSchema
from marshmallow import EXCLUDE, RAISE, Schema, fields, validate
from marshmallow_utils.fields import IdentifierSet, SanitizedUnicode
from marshmallow_utils.schemas import IdentifierSchema

from ...services.schema import BaseVocabularySchema


class AffiliationSchema(BaseVocabularySchema):
    """Service schema for affiliations."""

    acronym = SanitizedUnicode()
    identifiers = IdentifierSet(fields.Nested(IdentifierSchema))
    name = SanitizedUnicode(required=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module for managing vocabularies."""

import os

from setuptools import find_packages, setup

readme = open("README.rst").read()
history = open("CHANGES.rst").read()

tests_require = [
    "invenio-app>=1.3.0",
    "pytest-invenio>=1.4.1",
]

# Should follow inveniosoftware/invenio versions
invenio_search_version = ">=1.4.1,<2.0.0"
invenio_db_version = ">=1.0.9,<2.0.0"

extras_require = {
    "docs": [
        "Sphinx>=3,<4",
    ],
    "elasticsearch6": [
        "invenio-search[elasticsearch6]{}".format(invenio_search_version),
    ],
    "elasticsearch7": [
        "invenio-search[elasticsearch7]{}".format(invenio_search_version),
    ],
    # Databases
    "mysql": [
        "invenio-db[mysql,versioning]{}".format(invenio_db_version),
    ],
    "postgresql": [
        "invenio-db[postgresql,versioning]{}".format(invenio_db_version),
    ],
    "sqlite": [
        "invenio-db[versioning]{}".format(invenio_db_version),
    ],
    "tests": tests_require,
}

all_requires = []
for key, reqs in extras_require.items():
    if key in {"elasticsearch6", "elasticsearch7"}:
        continue
    all_requires.extend(reqs)
extras_require["all"] = all_requires

setup_requires = [
    "Babel>=2.8",
]

install_requires = [
    "flask>=1.1,<2",
    "invenio-records-resources>=0.16.5,<0.17.0",
    "invenio-i18n>=1.3.0",
]

packages = find_packages()

# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join("invenio_vocabularies", "version.py"), "rt") as fp:
    exec(fp.read(), g)
    version = g["__version__"]

setup(
    name="invenio-vocabularies",
    version=version,
    description=__doc__,
    long_description=readme + "\n\n" + history,
    keywords="invenio vocabulary management",
    license="MIT",
    author="CERN",
    author_email="info@inveniosoftware.org",
    url="https://github.com/inveniosoftware/invenio-vocabularies",
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    entry_points={
        "flask.commands": [
            "vocabularies = invenio_vocabularies.cli:vocabularies",
        ],
        "invenio_base.apps": [
            "invenio_vocabularies = invenio_vocabularies:InvenioVocabularies",
        ],
        "invenio_base.api_apps": [
            "invenio_vocabularies = invenio_vocabularies:InvenioVocabularies",
        ],
        "invenio_base.api_blueprints": [
            'invenio_vocabularies = invenio_vocabularies.views:create_blueprint_from_app',
        ],
        "invenio_db.alembic": [
            "invenio_vocabularies = invenio_vocabularies:alembic",
        ],
        "invenio_db.models": [
            "vocabulary_model = invenio_vocabularies.records.models",
        ],
        "invenio_jsonschemas.schemas": [
            "jsonschemas = invenio_vocabularies.records.jsonschemas",
        ],
        "invenio_search.mappings": [
            "vocabularies = invenio_vocabularies.records.mappings",
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 3 - Alpha",
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary service schema."""

from invenio_records_resources.services.records.schema import BaseRecordSchema
from marshmallow import EXCLUDE, RAISE, Schema, fields, validate
from marshmallow_utils.fields import SanitizedUnicode

i18n_strings = fields.Dict(
    allow_none=False,
    keys=fields.Str(validate=validate.Regexp("^[a-z]{2}$")),
    values=fields.Str(),
)
"""Field definition for language aware strings."""


class BaseVocabularySchema(BaseRecordSchema):
    """Base schema for vocabularies."""

    title = i18n_strings
    description = i18n_strings
    icon = fields.Str(allow_none=False)


class VocabularySchema(BaseVocabularySchema):
    """Service schema for vocabulary records.."""

    props = fields.Dict(
        allow_none=False, keys=fields.Str(), values=fields.Str()
    )
    tags = fields.List(SanitizedUnicode())
    type = fields.Str(attribute='type.id', required=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations Marshmallow schema tests."""

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.contrib.affiliations.schema import AffiliationSchema


def test_valid_full(affiliation_full_data):
    loaded = AffiliationSchema().load(affiliation_full_data)
    assert affiliation_full_data == loaded


def test_valid_minimal():
    data = {
        "name": "Test affiliation",
    }
    loaded = AffiliationSchema().load(data)
    assert data == loaded


def test_invalid_no_name():
    invalid = {
        "acronym": "TEST",
        "id": "aff-1",
        "identifiers": [
            {"identifier": "03yrm5c26", "scheme": "ror"}
        ],
        "title": {
            "en": "Test affiliation",
            "es": "Afiliacion de test"
        }
    }
    with pytest.raises(ValidationError):
        data = AffiliationSchema().load(invalid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Records-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""System field context for the Vocabulary PID field.

The context overrides the PID resolver to be aware of the vocabulary type, and
hence the PID type.

The context is used when you initialise a PIField, for instance:

.. code-block:: python

    class Vocabulary(Record):
        pid = PIDField(
            'id',
            provider=VocabularyIdProvider,
            context_cls=VocabularyPIDFieldContext
        )

You can then resolve vocabulary records using the type:

.. code-block:: python

    Vocabulary.pid.resolve(('<type>', '<pid_value>'))


Also, it's possible to make initialise the field with a type context:

.. code-block:: python

    Vocabulary.pid.with_type_ctx('<type>').resolve('<pid_value>')

"""

from copy import copy

from invenio_db import db
from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_records_resources.records.systemfields.pid import PIDFieldContext

from invenio_vocabularies.records.models import VocabularyType


class BaseVocabularyPIDFieldContext(PIDFieldContext):
    """PIDField context for vocabularies.

    This class implements the class-level methods available on a PIDField
    for vocabulary records.
    """

    def create(self, record):
        """Proxy to the field's create method."""
        return self.field.create(record)

    def delete(self, record):
        """Proxy to the field's delete method."""
        return self.field.delete(record)


class VocabularyPIDFieldContext(BaseVocabularyPIDFieldContext):
    """PIDField context for vocabularies.

    This class implements the class-level methods available on a PIDField
    for vocabulary records.
    """

    def resolve(self, pid_value):
        """Resolve identifier.

        :params pid_value: Either a tuple ``(type_id, pid_value)`` or just a
            ``pid_value`` if the type context has been initialized using
            ``with_type_ctx()``.
        """
        pid_type = self.pid_type
        if pid_type is None:
            type_id, pid_value = pid_value
            pid_type = self.get_pid_type(type_id)

        # Create resolver
        resolver = self.field._resolver_cls(
            pid_type=pid_type,
            object_type=self.field._object_type,
            getter=self.record_cls.get_record
        )

        # Resolve
        pid, record = resolver.resolve(pid_value)

        # Store pid in cache on record.
        self.field._set_cache(record, pid)

        return record

    def get_pid_type(self, type_id):
        """Get the PID type for a vocabulary type."""
        # Get type based on name.
        vocab_type = VocabularyType.query.filter_by(id=type_id).one_or_none()
        if vocab_type is None:
            raise PIDDoesNotExistError(None, None)
        return vocab_type.pid_type

    @property
    def pid_type(self):
        """Get the current defined type."""
        # This ensures that when we use Vocabulary.pid.with_type_ctx('...')
        # we cache the pid type to avoid querying the database every time.
        type_id = getattr(self, '_type_id', None)
        if type_id:
            pid_type = getattr(self, '_pid_type', None)
            if pid_type is None:
                pid_type = self.get_pid_type(type_id)
                self._pid_type = pid_type
            return pid_type

    def with_type_ctx(self, type_id):
        """Returns a new context initialized with the type context."""
        ctx = copy(self)
        ctx._type_id = type_id
        return ctx



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary resource."""

from functools import wraps

import marshmallow as ma
from flask import g
from flask_resources import JSONSerializer, MarshmallowJSONSerializer, \
    ResponseHandler, resource_requestctx, response_handler
from invenio_records_resources.resources import RecordResource, \
    RecordResourceConfig, SearchRequestArgsSchema
from invenio_records_resources.resources.records.headers import etag_headers
from invenio_records_resources.resources.records.resource import \
    request_data, request_headers, request_search_args, request_view_args
from invenio_records_resources.resources.records.utils import es_preference
from marshmallow import fields

from .serializer import VocabularyL10NItemSchema, VocabularyL10NListSchema


#
# Request args
#
class VocabularySearchRequestArgsSchema(SearchRequestArgsSchema):
    """Add parameter to parse tags."""

    tags = fields.Str()


#
# Resource config
#
class VocabulariesResourceConfig(RecordResourceConfig):
    """Vocabulary resource configuration."""

    blueprint_name = "vocabularies"
    url_prefix = "/vocabularies"
    routes = {
        "list": "/<type>",
        "item": "/<type>/<pid_value>",
    }

    request_view_args = {
        "pid_value": ma.fields.Str(),
        "type": ma.fields.Str(required=True),
    }

    request_search_args = VocabularySearchRequestArgsSchema

    response_handlers = {
        "application/json": ResponseHandler(
            JSONSerializer(),
            headers=etag_headers
        ),
        "application/vnd.inveniordm.v1+json": ResponseHandler(
            MarshmallowJSONSerializer(
                schema_cls=VocabularyL10NItemSchema,
                many_schema_cls=VocabularyL10NListSchema,
            ),
            headers=etag_headers,
        ),
    }


#
# Resource
#
class VocabulariesResource(RecordResource):
    """Resource for generic vocabularies."""

    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def search(self):
        """Perform a search over the items."""
        hits = self.service.search(
            identity=g.identity,
            params=resource_requestctx.args,
            type=resource_requestctx.view_args["type"],
            es_preference=es_preference(),
        )
        return hits.to_dict(), 200

    @request_view_args
    @request_data
    @response_handler()
    def create(self):
        """Create an item."""
        item = self.service.create(
            g.identity,
            resource_requestctx.data or {},
        )
        return item.to_dict(), 201

    @request_view_args
    @response_handler()
    def read(self):
        """Read an item."""
        pid_value = (
            resource_requestctx.view_args["type"],
            resource_requestctx.view_args["pid_value"]
        )
        item = self.service.read(pid_value, g.identity)
        return item.to_dict(), 200

    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def update(self):
        """Update an item."""
        pid_value = (
            resource_requestctx.view_args["type"],
            resource_requestctx.view_args["pid_value"]
        )
        item = self.service.update(
            pid_value,
            g.identity,
            resource_requestctx.data,
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return item.to_dict(), 200

    @request_headers
    @request_view_args
    def delete(self):
        """Delete an item."""
        pid_value = (
            resource_requestctx.view_args["type"],
            resource_requestctx.view_args["pid_value"]
        )
        self.service.delete(
            pid_value,
            g.identity,
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return "", 204



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Version information for Invenio-Vocabularies.

This file is imported by ``invenio_vocabularies.__init__``,
and parsed by ``setup.py``.
"""

__version__ = "0.7.4"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary facets."""

from flask_principal import AnonymousIdentity
from invenio_i18n.ext import current_i18n
from marshmallow_utils.fields.babel import gettext_from_dict
from speaklater import make_lazy_string

from ..proxies import current_service


def lazy_get_label(vocab_item):
    """Lazy evaluation of a localized vocabulary label."""
    params = {
        "locale": current_i18n.locale,
        "default_locale": "en"
    }

    return make_lazy_string(
        gettext_from_dict, vocab_item, **params)


class VocabularyLabels:
    """Fetching of vocabulary labels for facets."""

    def __init__(self, vocabulary, cache=False):
        """Initialize the labels."""
        self.vocabulary = vocabulary
        self.cache = cache
        self.fields = ["id", "title"]  # not configurable

    def __call__(self, ids):
        """Return the mapping when evaluated."""
        identity = AnonymousIdentity()
        if not self.cache:
            vocabs = current_service.read_many(
                identity, type=self.vocabulary, ids=ids, fields=self.fields)
        else:
            vocabs = current_service.read_all(
                identity, type=self.vocabulary, fields=self.fields)

        labels = {}
        vocab_list = list(vocabs.hits)  # the service returns a generator
        ids = set(ids)
        seen = set()
        for vocab in vocab_list:
            # cannot loop over ids because vocab id is inside each item
            if len(ids) == len(seen):
                break

            id_ = vocab["id"]
            if id_ in ids:
                labels[id_] = lazy_get_label(vocab["title"])
                seen.add(id_)

        return labels
